#!/usr/bin/env python3

## Plugin ideas

# `figure-tif`: A directive that's just like figure, but it uses imagemagick to 
# convert its image from a TIF to a PNG, then copies the PNG to the output 
# directory.

# A way to display a command and the image/plot it generates.

# A directive for listing reagents and product numbers.

## Add an OD600 directive

from docutils import nodes
from docutils.parsers.rst.directives.tables import Table

def dilution(argument): #
    import re
    match = re.match(r'(\d+)x', str(argument))

    if match:
        return int(match.group(1))
    else:
        raise ValueError(f"dilution should be a number followed by 'x' (e.g. '20x'), not '{argument}'")

def make_row(*cells): #
    row = nodes.row()

    for text in map(str, cells):
        entry = nodes.entry()
        entry += [nodes.paragraph(text, text)]
        row += entry

    return row

def make_cols(directive, num_cols): #
    tgroup = nodes.tgroup(cols=num_cols)
    col_widths = directive.get_column_widths(num_cols)

    for w in col_widths:
        colspec = nodes.colspec()
        colspec['colwidth'] = w
        tgroup += colspec

    return tgroup

def align_cols(table, *alignments): #
    for tgroup in table:
        if isinstance(tgroup, nodes.tgroup):
            for child in tgroup:
                if isinstance(child, (nodes.thead, nodes.tbody)):
                    for row in child:
                        for cell, alignment in zip(row, alignments):
                            cell['classes'].append(f'{alignment}-align')

class OdDirective(Table): #
    has_content = True
    option_spec = {
            'title': lambda x: x,
            'dilution': dilution,
            **Table.option_spec
    }
    
    def run(self): #
        table = nodes.table()

        dilution = self.options.get('dilution', None)
        num_cols = 3 if dilution else 2
        col_widths = self.get_column_widths(num_cols)

        tgroup = nodes.tgroup(cols=num_cols)
        table += tgroup

        for w in col_widths:
            colspec = nodes.colspec()
            colspec['colwidth'] = w
            tgroup += colspec

        thead = nodes.thead()
        title = self.options.get('title', 'Strain')

        if dilution:
            thead += make_row(title, f'OD600/{dilution}', 'OD600')
        else:
            thead += make_row(title, 'OD600')

        tbody = nodes.tbody()
        for line in self.content:
            strain, od = line.split(':')

            if dilution:
                real_od = f'{float(dilution) * float(od):0.3f}'
                tbody += make_row(strain, od, real_od)
            else:
                tbody += make_row(strain, od)

        tgroup += thead
        tgroup += tbody

        alignments = ['left'] + (num_cols - 1) * ['right']
        align_cols(table, *alignments)

        return [table]

## Add an electrotransformation directive

# I should rename this directive, now that I'm doing library-scale chemical 
# transformations with yeast.

# .. electrotransformation::
#    :dilutions: 10 ^ 1 2 3 4
#    :plate: 20 μL
# 
#    theo rhi 4,4:
#       colonies: x x 69 5
#       recover: 2032 μL
#       conc: 115.8 ng/μL
#    theo rhi 4,5:
#       colonies: x x 54 10
#       recover: 3070 μL
#       conc: 100.0 ng/μL
#    theo rhi 5,4:
#       colonies: x x 85 7
#       recover: 3057 μL
#       conc: 147.9 ng/μL
#    theo rhi 5,5:
#       colonies: x x 81 10
#       recover: 4089 μL
#       conc: 137.1 ng/μL

# .. electrotransformation::
#    :dilutions: 5 ^ 1 2 3
#    :plate: 20 μL
# 
#    theo ux:
#       colonies: 17 9 2
#       recover: 20800 μL
#       library_size: 1024 + 16384 + 262144

from docutils import nodes
from docutils.parsers.rst.directives.tables import Table
from nonstdlib import sci

class Dilutions: #

    def __init__(self, *args): #
        self.base = None
        self.powers = None
        self.override = None

        if len(args) == 1:
            self.override = args[0]
        elif len(args) == 2:
            self.base, self.powers = args
        else:
            raise ValueError(f"expected 1 or 2 arguments, got {len(args)}.")

    @property #
    def dilutions(self):
        if self.override:
            return self.override
        else:
            return [self.base ** x for x in self.powers]

    @property #
    def labels(self):
        if self.override:
            format = lambda x: sci(x)
        else:
            format = lambda x: sci(x, base=self.base, exp_only=True)

        return [format(x) for x in self.dilutions]

def dilution_list(argument): #
    tokens = argument.split()

    # If the dilution is specified as something like "10 ^ 2 3 4 5", calculate 
    # the dilutions as 10**3, 10**4, etc. and keep note of the fact that all 
    # the dilutions are powers of 10.  This will allow us to make more succinct 
    # column headers later on.
    if len(tokens) > 2 and tokens[1] == '^':
        base = eval(tokens[0])
        powers = list(map(eval, tokens[2:]))
        return Dilutions(base, powers)

    # Otherwise, treat each word as a different dilution.
    else:
        dilutions = map(eval, tokens)
        return Dilutions(dilutions)

def value_in(unit, cast=float): #

    def converter(argument): #
        import re
        unit.replace('μ', '[uμ]')
        value = r'([-+]?[0-9]*\.?[0-9]+)'
        match = re.match(f'({value}) {unit}', str(argument))

        if match:
            return cast(match.group(1))
        else:
            raise ValueError(f"expected a value in {unit} (e.g. '20 {unit}'), not '{argument}'")

    return converter

# Fucks up if wrong number of colony counts are given (shifts subsequent 
# columns over, should either fill with zero or throw error).
#
# Throws exceptions in a few places where it should use directive.error().

class ElectrotransformationDirective(Table): #
    has_content = True
    option_spec = {
            'dilutions': dilution_list,
            'plate': value_in('μL'),
            'recover': value_in('μL'),
            **Table.option_spec
    }
    
    def run(self): #
        data = self.load_data()
        dilutions = self.options.get('dilutions', dilution_list('10 ^ 2 3 4 5'))

        headers = [
                'Library',
                'DNA (ng/μL)',
                'Plated (μL)',
                'Recovered (μL)',
                *[f'CFU/{x}' for x in dilutions.labels],
                '# Transformed',
                '# Unique',
                'Coverage (%)',
                'Coverage (fold)',
        ]

        table = nodes.table()

        num_cols = len(headers)
        tgroup = make_cols(self, num_cols)
        table += tgroup

        thead = nodes.thead()
        thead += make_row(*headers)
        tgroup += thead

        import sys; sys.path.append('/home/kale/sgrna/scripts')
        from count_transformants import count_transformants, parse_library, evaluate_coverage

        tbody = nodes.tbody()
        tgroup += tbody

        for lib in data:
            plate_uL = data[lib]['plate']
            recover_uL = data[lib]['recover']
            num_cfus = data[lib]['colonies']
            num_unique = data[lib].get('library_size')
            if num_unique is not None and not isinstance(num_unique, int):
                num_unique = eval(num_unique)
            num_transformed = count_transformants(
                    recover_uL, plate_uL, num_cfus,
                    dilutions=dilutions.dilutions)
            library_name, library_size = parse_library(lib, num_unique)
            num_unique, percent_coverage, fold_coverage = evaluate_coverage(
                    library_size, num_transformed)
                
            assert len(dilutions.labels) == len(num_cfus)

            tbody += make_row(
                    lib,
                    data[lib].get('conc', '—'),
                    f'{plate_uL:.0f}',
                    f'{recover_uL:.0f}',
                    *num_cfus,
                    f'{sci(num_transformed)}',
                    f'{int(round(num_unique))}',
                    f'{percent_coverage * 100:.1f}%',
                    f'{fold_coverage:.1f}x',
            )

        alignments = ['left'] + (num_cols - 1) * ['right']
        align_cols(table, *alignments)

        return [table]

    def load_data(self): #
        """
        Parse the content of the directive using YAML, then cast things and 
        fill in default values as necessary.
        """
        import yaml
        data = yaml.load('\n'.join(self.content))

        default_plate = self.options.get('plate', 20)
        default_recover = self.options.get('recover', 1100)

        def colonies(argument):
            colonies = []
            for x in argument.split():
                if x in 'xX-*': colonies.append('lawn')
                else: colonies.append(eval(x))
            return colonies

        for lib in data:
            if 'colonies' not in data[lib]:
                raise ValueError(f"'colonies' not specified for '{lib}'")
                
            data[lib]['colonies'] = colonies(data[lib]['colonies'])

            if 'plate' in data[lib]:
                data[lib]['plate'] = value_in('μL')(data[lib]['plate'])
            else:
                data[lib]['plate'] = default_plate

            if 'recover' in data[lib]:
                data[lib]['recover'] = value_in('μL')(data[lib]['recover'])
            else:
                data[lib]['recover'] = default_recover

            if 'conc' in data[lib]:
                data[lib]['conc'] = value_in('ng/μL')(data[lib]['conc'])

        return data

def setup(app):
    # Add some science-themed directives.
    app.add_directive('od', OdDirective)
    app.add_directive('electrotransformation', ElectrotransformationDirective)




#!/usr/bin/env python3

from . import cli
from .. import Workspace

@cli.priority(40)
def init():
    """\
    Create a directory layout for a new project.  This layout will include 
    directories for data files, analysis scripts, protocols, notebook entries, 
    and documents.

    Usage:
        exmemo [project] init <title>

    Argument:
        <title>
            The title of the project.  This will be used to make nicely 
            formatted headers in various places, so it should be title-cased 
            and the words should be separated by spaces, e.g.

                "Anglo-French Silly Walk"

            Note that you have to quote the title, otherwise the shell will 
            interpret it as multiple arguments.

    This command will create the following directory structure:

    analysis/
       A python package containing code used for data analysis.  By virtue of 
       being a package, this code can be run from anywhere in the project.  
       You'll need to install it before you can use it:

          $ pip install -e analysis

    data/
       A single directory for any data file you collect.  Data files that are 
       relevant to a particular experiment can be symlinked into the notebook 
       directory using `exmemo data link`.  By  default, this directory is 
       ignored by git.

    documents/
       A directory for presentations, papers, and the like.  By default, this 
       directory is ignored by git.

    notebook/
       A Sphinx directory with separate directories for each experiment you 
       carry out.  An experiment is defined by asking and answering a single 
       question, which may take a single day or several months.

    protocols/
       A directory for protocols, which can be text files, python script, Excel 
       files, Word documents, whatever.
            
    Exmemo is fairly opinionated about which files to create and what to call 
    them, but for the most part these opinions are arbitrary and you can change 
    any you don't like.  The only things you shouldn't change are the names of 
    the directories in the project root.

    There are two ways to override exmemo's opinions.  The quick and easy way 
    is just to manually rename or edit any files.  The more thorough way is to 
    create a custom cookiecutter for your new projects and to tell exmemo about 
    it.  Cookiecutter (https://github.com/audreyr/cookiecutter) is a tool that 
    create new directory layouts from a template.  You can override exmemo's 
    default template by setting the 'cookiecutter_url' in one of the config 
    files, for example `~/.config/exmemo/conf.toml`:

        cookiecutter_url = https://github.com/yourname/cookiecutter-exmemo.git
    """
    args = cli.parse_args_via_docopt()
    work = Workspace('.')

    work.init_project(args['<title>'])

def root():
    """
    Print the root directory of the project.

    Usage:
        exmemo project root
    """
    args = cli.parse_args_via_docopt()
    work = Workspace.from_cwd()

    print(work.root_dir)





#!/usr/bin/env python3

import inspect
from pkg_resources import iter_entry_points

def get_plugins(group):
    plugins = []

    for entry_point in iter_entry_points(group=group):
        plugin = entry_point.load()
        plugin.name = entry_point.name
        plugin.module = entry_point.module_name
        plugin.lineno = inspect.getsourcelines(plugin)[1]
        plugin.priority = getattr(plugin, 'priority', 0)
        plugins.append(plugin)

    plugins.sort(key=lambda x: x.lineno)
    plugins.sort(key=lambda x: x.priority, reverse=True)

    return plugins



#!/usr/bin/env python3

import os
import subprocess

# Make the "documents", "data", and "protocols" directories.  These can't be 
# included directly in the cookiecutter, because they're empty, so can't be 
# part of the actual repository.

os.makedirs('documents', exist_ok=True)
os.makedirs('data', exist_ok=True)
os.makedirs('protocols', exist_ok=True)
os.makedirs('notebook/build', exist_ok=True)

# Make a git repository for the project and make the initial commit.

subprocess.run(['git', 'init'])
subprocess.run(['git', 'add', '.'])
subprocess.run(['git', 'commit', '-m', "Initial commit."])



#!/usr/bin/env python3

import toml
from . import cli
from .. import Workspace
from ast import literal_eval

def get():
    """\
    Get the value of a configuration option.

    Usage:
        exmemo config get <option>
    """
    args = cli.parse_args_via_docopt()
    work = Workspace.from_cwd(strict=False)
    option = args['<option>']

    if option in work.config:
        print(f"{option}: {work.config[option]}")
    else:
        print(f"Option '{option}' is not defined.")

def set():
    """\
    Set a configuration option.

    Usage:
        exmemo config set <option> <value> [--user | --site]

    Options:
        -s --site
            Set the given option in the site-wide configuration file.

        -u --user
            Set the given option in the user-wide configuration file.  This
            is the default behavior if the command is not run from within an 
            exmemo project directory.

    An effort will be made to interpret the given value as a python data type 
    (using `ast.literal_eval`, to be specific), but if this fails the value 
    will be interpreted as a string.  You can put quotes around a value 
    (usually two sets of quotes, since the first will be interpreted by the 
    shell) to force it to be interpreted as a string (for example, if you want 
    to store the literal string "42").
    """
    args = cli.parse_args_via_docopt()
    option, value = args['<option>'], args['<value>']
    work, path = get_config_path(args)

    # Try to interpret the given value as a python data type.
    try: value = literal_eval(value)
    except ValueError: pass

    # Update the config.
    config = toml.load(str(path))
    config[option] = value
    with open(path, 'w') as file:
        toml.dump(config, file)

def edit():
    """\
    Open the configuration file in your editor.

    Usage:
        exmemo config edit [--user | --site]

    Options:
        -s --site
            Open the site-wide configuration file, rather than the one for 
            this project.

        -u --user
            Open the site-wide configuration file, rather than the one for 
            this project.  This is the default behavior if the command is not 
            run from within an exmemo project directory.
    """
    args = cli.parse_args_via_docopt()
    work, path = get_config_path(args)
    work.launch_editor(path)


def get_config_path(args):
    """
    Return the path to the config file indicated by the arguments, along with a 
    workspace.

    This is used by the functions that can edit the config files.  Briefly, if 
    either the '--site' or '--user' option are specified, return the path to 
    the corresponding configuration file.  Otherwise, return the path to the 
    in-project configuration file if the current working directory is in a 
    project, or the user-wide configuration file if it isn't.
    """
    strict = not any((args['--site'], args['--user']))
    work = Workspace.from_cwd(strict=strict)

    if args['--site']:
        path = work.site_config_path
    elif args['--user'] or not work.has_project_files:
        path = work.user_config_path
    else:
        path = work.project_config_path

    return work, path



#!/usr/bin/env python3

import sys
from . import cli
from .. import Workspace, readers, utils

known_extensions = '\n    '.join(
        ' '.join(Reader.extensions)
        for Reader in readers.get_readers()
)
shared_doc = f"""\
Exmemo looks for protocols both inside and outside of the project.  Inside the 
project, exmemo looks in the current working directory and the `protocols/` 
directory (and all of its subdirectories) at the root of the project.  Outside 
the project, exmemo looks in any directory listed in the `shared_protocols` 
configuration option.  See `exmemo config -h` for information on setting 
configuration options.  This makes it easy to have protocols that are shared 
between projects:

    shared_protocols = ['~/research/protocols']

Only protocols with a recognized file extension will be displayed.  
Currently, the following extensions are recognized:
    
    {known_extensions}

Post a bug report if you have protocols of a different type that you would 
like to use with Exmemo.
"""

@cli.priority(20)
def show():
    """\
    Display the given protocol.

    Usage:
        exmemo [protocol] show <substr> [<args>...]

    Arguments:
        <substr>
            A string specifying the protocol to show.  You can provide any 
            substring from the name of the protocol.  If the substring is not 
            unique, you'll be asked which file you meant.

        <args>
            Any information that needs to be passed to the protocol.  This is 
            mostly useful for protocols that are scripts.

    {shared_doc}
    """
    argv = pop_protocol_args('show')
    args = cli.parse_args_via_docopt(shared_doc=shared_doc)
    work = Workspace.from_cwd(strict=False)
    protocol = work.pick_protocol(args['<substr>'])
    reader = readers.pick_reader(protocol, argv)

    reader.show(work)

@cli.priority(20)
def print():
    """\
    Print the given protocol.

    Usage:
        exmemo [protocol] print <substr>

    Arguments:
        <substr>
            A string specifying the protocol to print.  You can provide any 
            substring from the name of the protocol.  If the substring is not 
            unique, you'll be asked which file you meant.

    {shared_doc}
    """
    argv = pop_protocol_args('print')
    args = cli.parse_args_via_docopt(shared_doc=shared_doc)
    work = Workspace.from_cwd(strict=False)
    protocol = work.pick_protocol(args['<substr>'])
    reader = readers.pick_reader(protocol, argv)

    reader.print(work)
    reader.archive(work)

@cli.priority(20)
def archive():
    """\
    Save the protocol to a date-stamped text file that can be included in your 
    lab notebook.

    Usage:
        exmemo [protocol] archive <substr> [<dir>]

    Arguments:
        <substr>
            A string specifying the protocol to archive.  You can provide any 
            substring from the name of the protocol.  If the substring is not 
            unique, you'll be asked which file you meant.

        <dir>
            The directory to save the archive in.  By default this is just the 
            current working directory.

    {shared_doc}
    """
    argv = pop_protocol_args('archive')
    args = cli.parse_args_via_docopt(shared_doc=shared_doc)
    work = Workspace.from_cwd(strict=False)
    protocol = work.pick_protocol(args['<substr>'])
    reader = readers.pick_reader(protocol, argv)

    reader.archive(work, args['<dir>'])

def edit():
    """
    Edit the given protocol.

    Usage:
        exmemo protocol edit <substr>

    Arguments:
        <substr>
            A string specifying the protocol to show.  You can provide any 
            substring from the name of the protocol.  If the substring is not 
            unique, you'll be asked which file you meant.

    {shared_doc}
    """
    args = cli.parse_args_via_docopt(shared_doc=shared_doc)
    work = Workspace.from_cwd(strict=False)
    protocol = work.pick_protocol(args['<substr>'])
    reader = readers.pick_reader(protocol, [])

    reader.edit(work)

def ls():
    """\
    List protocols.

    Usage:
        exmemo protocol ls [<substr>]

    Arguments:
        <substr>
            Only list files that contain the given substring.

    {shared_doc}
    """
    args = cli.parse_args_via_docopt(shared_doc=shared_doc)
    work = Workspace.from_cwd(strict=False)

    for last, dir in utils.last(work.protocols_dirs):
        print(dir)
        for path in work.iter_protocols_from_dir(dir, args['<substr>']):
            print(' ', path.relative_to(dir))
        if not last:
            print()

def plugins():
    """
    List every installed plugin for reading protocols.

    Usage:
        exmemo protocol plugins

    You can create plugins to add support for file extensions that aren't 
    understood by Exmemo natively.  A plugin is a class, usually pretty simple, 
    with methods describing how to show, edit, print, and save a protocol of a 
    particular file type.  Look in ``exmemo/reader.py`` for examples of how to 
    write such a class, and ``setup.py`` for examples of how to install them.
    """
    args = cli.parse_args_via_docopt()
    for plugin in readers.get_readers():
        print(f"{plugin.name}: {' '.join(plugin.extensions)}")


def pop_protocol_args(subcommand):
    cut = sys.argv.index(subcommand) + 2
    sys.argv, protocol_args = sys.argv[:cut], sys.argv[cut:]
    return protocol_args



#!/usr/bin/env python3

from . import cli
from .. import Workspace

@cli.priority(30)
def new():
    """\
    Create a new experiment with a blank notebook entry.

    Usage:
        exmemo [note] new <title>

    Arguments:
        <title>
            The title of the experiment.  This should be title cased, with 
            words separated by spaces.  Use quotes so the shell won't interpret 
            it as multiple arguments.
    """
    args = cli.parse_args_via_docopt()
    work = Workspace.from_cwd()

    work.init_experiment(args['<title>'])

@cli.priority(30)
def edit():
    """\
    Open the notebook entry for the given experiment in a text editor.

    Usage:
        exmemo [note] edit [<substr>]

    Arguments:
        <substr>
            The experiment to edit.  You don't need to specify the whole name, 
            just enough to be unique.  If multiple experiments match the given 
            substr, you'll will be asked which one you mean.  If you don't 
            specify an experiment, the most recently created one will be used.

    You can specify which text editor you prefer in an `.exmemorc` file, or via 
    the $EDITOR environment variable.
    """
    args = cli.parse_args_via_docopt()
    work = Workspace.from_cwd()
    path = work.pick_notebook_entry(args['<substr>'])

    work.launch_editor(path)

@cli.priority(30)
def open():
    """\
    Open a new terminal cd'd into the given experiment.

    Usage:
        exmemo [note] open [<substr>]

    Arguments:
        <substr>
            The experiment to open.  You don't need to specify the whole name, 
            just enough to be unique.  If multiple experiments match the given 
            substr, you'll will be asked which one you mean.  If you don't 
            specify an experiment, the most recently created one will be used.

    You can specify which terminal you prefer in an `.exmemorc` file, or via 
    the $TERMINAL environment variable.
    """
    args = cli.parse_args_via_docopt()
    work = Workspace.from_cwd()
    expt = work.pick_experiment(args['<substr>'])

    work.launch_terminal(expt)

@cli.priority(30)
def directory():
    """\
    Print the path to the given experiment.  This is often used to create an 
    alias that cd's into the given experiment.

    Usage:
        exmemo note directory [<substr>]

    Arguments:
        <substr>
            The experiment you're interested in.  You don't need to specify the 
            whole name, just enough to be unique.  If multiple experiments 
            match the given substr, you'll will be asked which one you mean.  
            If you don't specify an experiment, the most recently created one 
            will be implied.

    Programs are not allowed to change the state of the shell (i.e. to move you 
    into a different directory), so if you want a command that automatically 
    cd's into a given experiment, you need to write your own shell function.  
    This is how such a function would look in bash:

        function ed () {
            d=$(exmemo note directory "$@")
            # Don't try to cd if something goes wrong.
            [ $? = 0 ] && cd $d || echo $d
        }
    """
    args = cli.parse_args_via_docopt()
    work = Workspace.from_cwd()
    expt = work.pick_experiment(args['<substr>'])

    print(expt.resolve())

@cli.priority(30)
def rename():
    """\
    Rename 

    Usage:
        exmemo note rename [<substr>] <new_title>
    """
    pass

@cli.priority(30)
def build():
    """\
    Render the lab notebook to HTML using Sphinx.

    Usage:
        exmemo [note] build [-f]

    Options:
        -f --force
            Force Sphinx to rebuild the whole project from scratch, rather than 
            just building the files that changed since last time.  You often 
            need to do this after you add a new experiment.
    """
    args = cli.parse_args_via_docopt()
    work = Workspace.from_cwd()

    work.build_notebook(args['--force'])

@cli.priority(30)
def browse():
    """
    Open the rendered lab notebook in a web browser.

    Usage:
        exmemo [note] browse [-w]

    Options:
        -w --new-window
            Open a new window in the browser.  The default is to just open a 
            new tab in whichever browser window you most recently used.

    The default browser is firefox, but you can change this by setting either 
    the `browser` configuration option or the `$BROWSER` environment variable.  
    If you plan to use the `-w` option with this command, you may also need to 
    specify the command-line switch used to get the browser to create a new 
    window.  The default is `--new-window`.  That works for both firefox and 
    chrome, but I don't know about other browsers.  You can change it by 
    setting either the `browser_new_window_flag` configuration option or the 
    `$BROWSER_NEW_WINDOW_FLAG` environment variable.
    """
    args = cli.parse_args_via_docopt()
    work = Workspace.from_cwd()
    url = work.notebook_dir / 'build' / 'html' / 'index.html'
    
    work.launch_browser(f'file://{url}', args['--new-window'])

def ls():
    """\
    Print the names of any existing experiments.

    Usage:
        exmemo note ls [<substr>]

    Arguments:
        <substr>
            Only print the experiments matching the given text.
    """
    args = cli.parse_args_via_docopt()
    workspace = Workspace.from_cwd()

    for expt in workspace.iter_experiments(args['<substr>']):
        print(expt.name)




#!/usr/bin/env python3

def ymd():
    from datetime import datetime
    return datetime.today().strftime('%Y%m%d')

def last(iterable):
    yield from (
            (i == len(iterable), x)
            for i, x in enumerate(iterable, 1)
    )

def pick_one(choices):
    # Print everything to sys.stderr, because stdout is often redirected.
    import sys, functools, builtins
    print = functools.partial(builtins.print, file=sys.stderr)
    input = lambda: print(end='> ') or builtins.input()

    print("Did you mean?")
    for i, value in enumerate(choices, 1):
        print(f"({i}) {value}")

    # Keep track of the number of choices here so we don't need to call `len()` 
    # later on.  This allows the choices argument to be an iterator.
    num_choices = i

    def is_input_ok(x):
        if x.lower() == 'q':
            raise EOFError

        try: x = int(x)
        except ValueError:
            return False

        if x < 1 or x > num_choices:
            return False

        return True

    choice = input()
    
    while not is_input_ok(choice):
        print(f"Please enter a number between 1 and {num_choices}.")
        choice = input()

    return int(choice) - 1



#!/usr/bin/env python3

from . import cli

@cli.main
def main():
    """\
    Manage a project directory.

    Usage:
        exmemo <command> [<args>...]
        exmemo (-h | --help)
        exmemo --version

    Commands:
        {subcommands}
    """
    cli.run_subcommand_via_docopt('exmemo.commands')

def project():
    """\
    Manage the entire project.

    Usage:
        exmemo project <command> [<args>...]
        exmemo project (-h | --help)
        exmemo project --version

    Commands:
        {subcommands}
    """

    cli.run_subcommand_via_docopt('exmemo.commands.project')

def note():
    """\
    Keep notes on your day-to-day experiments.

    Usage:
        exmemo note <command> [<args>...]
        exmemo note (-h | --help)
        exmemo note --version

    Commands:
        {subcommands}
    """
    cli.run_subcommand_via_docopt('exmemo.commands.note')

def protocol():
    """\
    Manage, display, and print protocols.

    Usage:
        exmemo protocol <command> [<args>...]
        exmemo protocol (-h | --help)
        exmemo protocol --version

    Commands:
        {subcommands}
    """
    cli.run_subcommand_via_docopt('exmemo.commands.protocol')

def data():
    """\
    Interact with data files.

    Usage:
        exmemo data <command> [<args>...]
        exmemo data (-h | --help)
        exmemo data --version

    Commands:
        {subcommands}
    """
    cli.run_subcommand_via_docopt('exmemo.commands.data')

def config():
    """\
    Get and set configuration options.

    Usage:
        exmemo config <command> [<args>...]
        exmemo config (-h | --help)
        exmemo config --version

    Commands:
        {subcommands}

    Each project reads configuration information from three different files.  
    Typical paths for a Linux system are given below:

        In-project: ~/path/to/project/.exmemorc
        User-wide: ~/.config/exmemo/conf.toml
        Site-wide: /etc/xdg/exmemo/conf.toml

    Values in the earlier files take precedence over those in later files.  All 
    three files are TOML (https://github.com/toml-lang/toml).  As of yet there 
    is no definitive list of every configuration option understood by exmemo, 
    but typically the help message for each command will mention any relevant 
    configuration options.
    """
    cli.run_subcommand_via_docopt('exmemo.commands.config')

def debug():
    """\
    Print information that can help diagnose problems with exmemo.

    Usage:
        exmemo debug <command> [<args>...]
        exmemo debug (-h | --help)
        exmemo debug --version

    Commands:
        {subcommands}
    """
    cli.run_subcommand_via_docopt('exmemo.commands.debug')




#!/usr/bin/env python3

import sys
import docopt
import textwrap
import functools
from .. import __version__, utils, get_plugins

def run_subcommand(group, name, level=None):
    """
    Run the subcommand with the given group and name.  This is cool because by 
    using `pkg_resources`, other packages can add commands to `exmemo`!
    """
    # Find all the commands that match what the user typed.
    known_subcommands = get_subcommands(group)
    matching_subcommands = [
            x for x in known_subcommands
            if x.name.startswith(name)
    ]

    # If no subcommands match, print an error message and exit.
    if len(matching_subcommands) == 0:
        raise UnknownSubcommand(group, name, known_subcommands)

    # If two or more subcommands match, ask the user which one they meant.
    if len(matching_subcommands) > 1:
        cmd = group.split('.'); del cmd[1]
        i = utils.pick_one(f'{" ".join(cmd)} {x.name}' for x in matching_subcommands)
        matching_subcommands = [matching_subcommands[i]]

    # By this point, there should only be one matching subcommand.
    assert len(matching_subcommands) == 1
    subcommand = matching_subcommands[0]

    # Put the full subcommand name in `sys.argv`, so the argument parser 
    # doesn't need to deal with the extra complexity of partial names.
    level = level or len(group.split('.')) - 1
    sys.argv[level] = subcommand.name

    subcommand()

def run_subcommand_via_docopt(group, level=None, doc=None, command='<command>', **format_args):
    doc = doc or get_caller_docstring()
    briefs = get_subcommand_briefs(group)
    level = level or len(group.split('.')) - 1

    args = docopt.docopt(
            doc=doc.format(subcommands=briefs, **format_args),
            argv=sys.argv[1:level+1],
            version=__version__,
    )

    if args[command]:
        run_subcommand(group, args[command])
        return True

    return False

def parse_args_via_docopt(**format_args):
    doc = get_caller_docstring()
    if format_args:
        doc = doc.format(**format_args)
    return docopt.docopt(doc)

def handle_docopt_help_with_pager(help, version, options, doc):
    """
    Monkey-patched version of `docopt.extras` that uses a pager (less) if the 
    help text is too big to fit on the screen.
    """
    from shutil import get_terminal_size
    from subprocess import run

    if help and any((o.name in ('-h', '--help')) and o.value for o in options):
        doc = doc.strip('\n')
        w, h = get_terminal_size()
        if doc.count('\n') > h: run('less', input=doc, encoding='utf8')
        else: print(doc)
        sys.exit()

    if version and any(o.name == '--version' and o.value for o in options):
        print(version)
        sys.exit()

docopt.extras = handle_docopt_help_with_pager


def get_subcommands(group):
    return get_plugins(group)

def get_subcommand_briefs(group):
    subcommands = get_subcommands(group)

    # Make sure each subcommand has a brief description.

    for subcmd in subcommands:
        if not hasattr(subcmd, 'brief'):

            # Extract a brief description from the docstring.
            subcmd.brief = get_docstring(subcmd).split('.')[0] + '.'

            # Complain if it looks like the docstring extraction didn't work.
            if not subcmd.brief or 'usage:' in subcmd.brief.lower():
                raise ValueError(f"No brief description for '{subcmd.module}.{subcmd.name}'")

    # Sort the subcommands.

    subcommands.sort(key=lambda x: x.lineno)
    subcommands.sort(key=lambda x: x.priority, reverse=True)
    
    # Format the brief descriptions

    sep, indent, padding = ':', 4, 1
    name_width = max(len(x.name) for x in subcommands) + len(sep)
    brief_width = 79 - name_width - indent - padding

    briefs = ""
    for subcmd in subcommands:
        briefs += (
                f"{' ' * indent}"
                f"{subcmd.name + sep:{name_width}s}"
                f"{' ' * padding}"
                f"{textwrap.shorten(subcmd.brief, brief_width)}"
                f"\n"
        )

    return briefs.strip()
    
def get_caller_docstring():
    import inspect
    frame_info = inspect.stack()[2]
    caller = frame_info.frame.f_globals[frame_info.function]
    return get_docstring(caller)

def get_docstring(func):
    return textwrap.dedent(func.__doc__ or '').strip()
    

def main(func):

    @functools.wraps(func)
    def decorator(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except (KeyboardInterrupt, EOFError):
            sys.exit(1)
        except Exception as e:
            if hasattr(e, 'show_message_and_die'):
                print(e.message)
                sys.exit(2)
            else:
                raise

    return decorator


def brief(desc, priority=None):
    def decorator(f):
        f.brief = brief
        if priority is not None:
            f.priority = priority
        return f
    return decorator

def priority(level):
    def decorator(f):
        f.priority = level
        return f
    return decorator


class UnknownSubcommand(Exception):
    show_message_and_die = True
    
    def __init__(self, group, name, known_subcommands):
        cmd = [*group.split('.'), name]; del cmd[1]
        self.message = f"""Unknown command '{" ".join(cmd)}'."""





#!/usr/bin/env python3

import os
import shlex
import subprocess
from pathlib import Path

def get_collectors():
    from .plugins import get_plugins
    return {x.name: x for x in get_plugins('exmemo.datacollectors')}

def sync_data(work, verbose):
    collectors = get_collectors()

    for config in work.config.get('data', []):
        try: type = config.pop('type')
        except KeyError:
            raise UnspecifiedCollectorType(config)

        try: collector_cls = collectors[type]
        except KeyError:
            raise UnknownCollectorType(type, collectors)

        collector = collector_cls(**config)
        collector.sync(work, verbose)

def run(cmd, verbose=True, **kwargs):
    if cmd is None:
        return
    if verbose:
        cmd_str = cmd if isinstance(cmd, str) else ' '.join(str(x) for x in cmd)
        print(f'$ {cmd_str}')
    return subprocess.run(cmd, **kwargs)


class RsyncCollector:
    """
    Copy any files that have changed from the given (and possibly remote) 
    source to the given destination within the project data directory.  The 
    following options are recognized:

    src (required):
        The file or directory to copy, relative to the project data 
        directory.  If you want to specify a directory, remember that 
        trailing slashes are significant to rsync: 'dir' means copy dir 
        itself, 'dir/' means copy only the files inside dir.
       
    dest (default: '.'):
        Where to copy the files to, relative to the project data directory.
       
    cmd (default: 'rsync --archive --ignore-existing {src} {dest}'):
        The rsync command to execute.  {src} and {dest} will be replaced 
        with the values of the `src` and `dest` options, respectively.  The 
        command will be run from the project data directory.  The purpose 
        of this option is to allow you to pass rsync different flags, or 
        even to run a different command entirely.

    precmd (default: ''):
        A shell command to execute before running rsync.  The command will 
        be run from the project data directory.

    postcmd (default: ''):
        A shell command to execute after running rsync.  The command will 
        be run from the project data directory.
    """

    def __init__(self, src, dest=None, cmd=None, precmd=None, postcmd=None):
        # Use `os.path.expanduser()` for `src` because it doesn't clobber 
        # trailing slashes, which are significant to rsync.
        self.src = os.path.expanduser(src)
        self.dest = Path(dest or '.').expanduser()
        self.cmd = cmd or 'rsync --archive --ignore-existing {src} {dest}'
        self.precmd = precmd or ''
        self.postcmd = postcmd or ''

    def sync(self, work, verbose):
        dest = work.data_dir / self.dest
        paths = dict(src=self.src, dest=dest)
        rsync = [x.format(**paths) for x in shlex.split(self.cmd)]

        for precmd in self.precmd.split('\n'):
            run(precmd, verbose, cwd=work.data_dir, shell=True)

        # Shell-mode disabled to eliminate the possibility of getting 
        # confused by spaces/quotes/whatever in file names.
        run(rsync, verbose, cwd=work.data_dir, shell=False)

        for postcmd in self.postcmd.split('\n'):
            run(postcmd, verbose, cwd=work.data_dir, shell=True)
    

class UsbCollector(RsyncCollector):
    """
    Copy files from a USB drive into the project.  Rsync is used to 
    actually copy the files, so large directories can be efficiently copied 
    and all of the options that apply to the 'rsync' collector also apply 
    here.
    
    src (required):
        The file or directory to copy.  If you want to specify a directory, 
        remember that trailing slashes are significant to rsync: 'dir' 
        means copy dir itself, 'dir/' means copy only the files inside dir.
       
    dest (default: '.'):
        Where to copy the files to, relative to the project data directory.

    mountpoint (default: None):
        The directory where the USB drive is mounted.  If this is 
        specified, exmemo will attempt to automatically mount and unmount 
        the USB drive if it can't otherwise locate the files.  If this 
        option is not specified, the USB drive must be mounted before the 
        sync command is run.
       
    cmd (default: 'rsync --archive --ignore-existing {src} {dest}'):
        The rsync command to execute.  {src} and {dest} will be replaced 
        with the values of the `src` and `dest` options, respectively.  The 
        command will be run from the project data directory.  The purpose 
        of this option is to allow you to pass rsync different flags, or 
        even to run a different command entirely.

    precmd (default: ''):
        A shell command to execute before running rsync, but after the USB 
        drive has been mounted.  The command will be run from the project 
        data directory.
    

    postcmd (default: ''):
        A shell command to execute after running rsync, but before the USB 
        drive has been unmounted.  The command will be run from the project 
        data directory.
    .
    """

    def __init__(self, src, dest=None, mountpoint=None, rsync=None, precmd=None, postcmd=None):
        super().__init__(src, dest, rsync, precmd, postcmd)
        self.mountpoint = mountpoint and Path(mountpoint).expanduser()

    def sync(self, work, verbose):
        """
        Copy files from the given location on a USB drive into the project.
        """

        def is_mounted():
            cmd = 'mountpoint', '-q', self.mountpoint
            return subprocess.run(cmd).returncode == 0

        umount_when_done = False

        if not os.path.exists(self.src):

            # Give up if the source doesn't exist and we weren't told how to in
            # mount it.
            if not self.mountpoint:
                return

            if not is_mounted():
                cmd = 'mount', self.mountpoint
                err = run(cmd, verbose).returncode
                umount_when_done = (err == 0)

        if os.path.exists(self.src):
            super().sync(work, verbose)
        elif verbose:
            print(f"Can't find {self.src}.")

        if umount_when_done:
            cmd = 'umount', self.mountpoint
            err = run(cmd, verbose).returncode
            if not err:
                print("USB safe to remove.")



class BadCollectorConfig(Exception):
    show_message_and_die = True

    def __init__(self, config):
        key_value = '\n'.join(f'  {k}: {v}' for k,v in config.items())
        self.message = f"""\
The following data collector is missing a type:

{key_value}"""


class UnknownCollectorType(Exception):
    show_message_and_die = True

    def __init__(self, type, collectors):
        if not collectors:
            self.message = "No data collectors installed."
        else:
            did_you_mean = '\n'.join(f'  {x}' for x in collectors)
            self.message = f"""\
Unknown data collector '{type}'.  Did you mean:

{did_you_mean}"""



# Make the collector docstrings line up right when they're indented and 
# included in a usage string.
# vim: tw=75



#!/usr/bin/env python3

import re
import time
from .. import app, Workspace
from docutils import nodes
from docutils.parsers.rst import Directive
from sphinx.roles import XRefRole
from pathlib import Path

from html.parser import HTMLParser

def add_expts_to_toc(app, docname, source):
    """
    Update the main TOC with the list of all the experiments.

    The source argument is a list whose single element is the contents of the 
    source file. You can process the contents and replace this item to 
    implement source-level transformations.
    """
    if docname != 'index':
        return

    work = Workspace.from_dir(app.srcdir)
    entries = [
            str(work.get_notebook_entry(x).relative_to(app.srcdir))
            for x in sorted(work.iter_experiments())]

    source[0] = source[0].format(notebook_entries='\n   '.join(entries))

def add_dates_to_toc(app, doctree):
    """
    Prefix the TOC entries for each experiment with a date.

    The dates are added as literal text so that they will line up with each 
    other vertically.
    """
    docname = app.env.docname

    if docname == 'index':
        return

    try:
        ref = app.env.tocs[docname][0][0][0]
        match = re.match('(\d{8})_.*', docname)
    except:
        return

    if match:
        date, pad = match.group(1), ' '
        ref.insert(0, nodes.literal(date, date))
        ref.insert(1, nodes.Text(pad, pad))

def doi_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    """
    Insert a citation (including the authors, title, journal, and publication 
    date) given any valid PubMed identifier (i.e. PMID, PMCID, or DOI).

    For example:
        :pubmed:`10.1093/nar/gkw908`

    Arguments
    =========
    name:       The role name used in the document.
    rawtext:    The entire markup snippet, with role.
    text:       The text marked with the role.
    lineno:     The line number where rawtext appears in the input.
    inliner:    The inliner instance that called us.
    options:    Directive options for customization.
    content:    The directive content for customization.

    Returns
    =======
    A 2 part tuple containing a list of nodes to insert into the document and a 
    list of system messages.  Both are allowed to be empty.
    """

    # There are some inefficiencies due to this function being called once for 
    # each role.  Ideally we'd like to only load the cache once, and ask for 
    # all the DOIs we care about in a single request to the Crossref.  There's 
    # probably a way to do this, but it would require some coordination between 
    # roles.

    # Dealing with all the roles together might also give me a way to line them 
    # up in a <table>.  Maybe I want a bibliography directive or something...

    # Also, would be nice to have formatting in the citations (i.e. bold, 
    # italic, etc.) and to have a link to the actual paper (via the DOI).

    import json
    import habanero
    import requests

    # Try to cache queries.
    cache_path = Path(app.user_cache_dir) / 'crossref.json'
    cache = {}

    if cache_path.exists():
        with cache_path.open() as f:
            cache = json.load(f)

    if text in cache:
        meta = cache[text]

    else:
        # Look up the article metadata on CrossRef.
        try:
            crossref = habanero.Crossref(mailto='kale_kundert@hms.harvard.edu')
            meta = crossref.works(ids=[text])['message']

            # The reference list is big, and we don't care about it, so don't 
            # cache it.
            if 'reference' in meta:
                del meta['reference']

            cache[text] = meta
            cache_path.parent.mkdir(exist_ok=True)
            with cache_path.open('w') as f:
                json.dump(cache, f)

        # Give a useful warning if the article can't be found.
        except requests.exceptions.HTTPError as e:
            warning = inliner.reporter.warning(f"No matches found for DOI: {text}", line=lineno)
            p = nodes.paragraph(text, text)
            return [p], [warning]

        except requests.exceptions.ConnectionError as e:
            warning = inliner.reporter.warning("Couldn't connect to CrossRef, may be offline.", line=lineno)
            p = nodes.paragraph(text, text)
            return [p], [warning]

    # Make a paragraph containing the citation.

    def format_author(author):
        initials = ''.join(x[0] for x in author['given'].split())
        return f"{author['family']} {initials}"

    def format_authors(authors):
        authors = [format_author(x) for x in authors]

        if len(authors) == 1:
            return authors[0]
        elif len(authors) < 5:
            return ", ".join(x for x in authors[:-1]) + " & " + authors[-1]
        else:
            return f"{authors[0]} et al"

    def format_title(meta):
        return meta['title'][0]

    def format_journal_issue_date(meta):
        issue = ':'.join(
                meta[k] for k in ('volume', 'issue', 'page')
                if k in meta
        )
        if issue:
            issue += ' '

        return f"{meta['container-title'][0]} {issue}({meta['issued']['date-parts'][0][0]})"

    citation = (
            f"{format_authors(meta['author'])}. "
            f"{format_title(meta)}. "
            f"{format_journal_issue_date(meta)}. "
    )

    p = nodes.paragraph(citation, citation)
    return [p], []


class ExperimentRole(XRefRole):
    """
    Make a hyperlink to another experiment.
    
    For example:
        :expt:`20170329_test_multiple_spacers`
    """
    innernodeclass = nodes.inline

    def __init__(self):
        # Copying from sphinx/domains/std.py:489
        super().__init__(warn_dangling=True)

    def process_link(self, env, refnode, has_explicit_title, title, target): #
        """
        Called after parsing title and target text, and creating the
        reference node (given in *refnode*).  This method can alter the
        reference node and must return a new (or the same) ``(title, target)``
        tuple.
        """
        refnode['refdomain'] = 'std'
        refnode['reftype'] = 'doc'

        slug = target.split('_', 1)[1]
        link = f'/{target}/{slug}'

        return super().process_link(env, refnode,
                has_explicit_title, title, link)


class UpdateDirective(Directive):
    """
    Define a directive specifically for adding new information to existing 
    notebook entries.

    .. update:: Dec 6, 2017

        Based on feedback I got at the conference, I repeated the analysis 
        for this experiment using method X...

    This directive takes a date as a positional argument, and then any amount 
    of text in an indented block.  It will render a box containing the given 
    text with a header that reads "Update — {date}".
    """
    required_arguments = 1  # <date>
    has_content = True
    final_argument_whitespace = True

    def run(self): #
        # Create the root admonition node.
        content = '\n'.join(self.content)
        options = {'classes': ['note']}
        admonition = nodes.admonition(content, **options)

        # Create the title node.
        title = f"Update — {self.arguments[0]}"
        text_nodes, _ = self.state.inline_text(title, self.lineno)
        title = nodes.title(title, '', *text_nodes)
        title.source, title.line = \
                self.state_machine.get_source_and_line(self.lineno)
        admonition += title

        # Create the content node.
        self.state.nested_parse(self.content, self.content_offset, admonition)

        return [admonition]

class ShowNodesDirective(Directive):
    """
    Pretty print all the nodes in the restructured-text contained in this 
    directive.  This is only meant to be a tool for developing and debugging 
    new Sphinx extensions.

    .. show-nodes::

        .. figure:: path/to/image.png

            A schematic of the reaction setup.
    """
    has_content = True

    def run(self): #
        wrapper = nodes.paragraph()
        self.state.nested_parse(self.content, self.content_offset, wrapper)

        for node in wrapper.children:
            print(node.pformat())

        return wrapper.children


def setup(app):
    app.connect('source-read', add_expts_to_toc)
    app.connect('doctree-read', add_dates_to_toc)

    app.add_role('expt', ExperimentRole())
    app.add_role('doi', doi_role)

    app.add_directive('update', UpdateDirective)
    app.add_directive('show-nodes', ShowNodesDirective)

    css_path = Path(__file__).parent / 'tweaks.css'
    app.add_stylesheet(str(css_path))



#!/usr/bin/env python3

import os
import sys
import platform
import toml
import formic
import shlex
import subprocess
import collections
from subprocess import DEVNULL
from datetime import datetime
from pathlib import Path
from appdirs import AppDirs
from . import readers, utils

app = AppDirs('exmemo')

class Workspace:

    @classmethod
    def from_dir(cls, dir, strict=True):
        """
        Create a workspace object containing given directory (or the current 
        working directory, by default).  This involves descending the directory 
        hierarchy looking for the root of the project, which should contain a 
        characteristic set of files and directories.
        """

        dir = given_dir = Path(dir)
        work = given_work = Workspace(dir)

        while not work.has_project_files:
            dir = dir.parent
            if dir == Path('/'):
                if strict: raise WorkspaceNotFound(given_dir)
                else: return given_work
            else:
                work = Workspace(dir)

        return work

    @classmethod
    def from_cwd(cls, strict=True):
        """
        Create a workspace object containing given directory (or the current 
        working directory, by default).  This involves descending the directory 
        hierarchy looking for the root of the project, which should contain a 
        characteristic set of files and directories.
        """
        return cls.from_dir(cls.get_cwd(), strict)

    @staticmethod
    def get_cwd():
        """
        Return the current working directory, making an effort to keep the path 
        nice and short by not resolving symlinks.
        """
        # Most shells will set `$PWD` with the path the user sees the current 
        # working directory as, taking into account whichever symlinks they 
        # used to get there.  So this is the path we want to use, if it's 
        # available.  If it's not, fall back on `os.getcwd()`, which will give 
        # us the current working directory without any symlinks.

        # The logic is really complicated because I ran into a bug in xonsh 
        # where `os.getenv('PWD')` and `os.getcwd()` would actually return 
        # different paths!  To be robust against this kind of situation, the 
        # code now checks if the two paths differ and trusts `os.getcwd()` if 
        # they do.

        pretty_cwd = os.getenv('PWD')
        real_cwd = os.getcwd()

        if not pretty_cwd:
            return Path(real_cwd)
        if not os.path.samefile(pretty_cwd, real_cwd):
            return Path(real_cwd)

        return Path(pretty_cwd)


    def __init__(self, root):
        # Use `os.path.abspath()` instead of `Path.resolve()` to avoid 
        # resolving any symlinks in the path.
        self._root = Path(os.path.abspath(root))

        if not self.config_paths:
            self._config = {}
        else:
            self._config = toml.load([str(x) for x in self.config_paths])

    @property
    def root_dir(self):
        return self._root

    @property
    def config(self):
        return self._config

    @property
    def config_paths(self):
        paths = [self.site_config_path, self.user_config_path, self.rcfile]
        return [x for x in paths if x.exists()]

    @property
    def site_config_path(self):
        return Path(app.site_config_dir) / 'conf.toml'

    @property
    def user_config_path(self):
        return Path(app.user_config_dir) / 'conf.toml'

    @property
    def project_config_path(self):
        return self.root_dir / '.exmemorc'

    rcfile = project_config_path

    @property
    def analysis_dir(self):
        return self.root_dir / 'analysis'

    @property
    def data_dir(self):
        return self.root_dir / 'data'

    @property
    def documents_dir(self):
        return self.root_dir / 'documents'

    @property
    def notebook_dir(self):
        return self.root_dir / 'notebook'

    @property
    def current_experiment_entry(self):
        cwd = self.get_cwd().resolve()
        if self.notebook_dir not in cwd.parents:
            return None
        for expt in self.iter_experiment_entries():
            if cwd == expt.parent or expt.parent in cwd.parents:
                return expt

    @property
    def current_experiment_dir(self):
        entry = self.current_experiment_entry
        return entry and entry.parent

    @property
    def protocols_dir(self):
        return self.root_dir / 'protocols'

    @property
    def protocols_dirs(self):
        local_dirs = [self.current_experiment_dir, self.protocols_dir]
        shared_dirs = [Path(x).expanduser() for x in self.config.get('shared_protocols', [])]
        return [x for x in local_dirs + shared_dirs if x and x.exists()]

    @property
    def has_project_files(self):
        if not self.rcfile.exists():
            return False
        if not self.analysis_dir.exists():
            return False
        if not self.data_dir.exists():
            return False
        if not self.documents_dir.exists():
            return False
        if not self.notebook_dir.exists():
            return False
        if not self.protocols_dir.exists():
            return False
        return True

    def iter_data(self, substr=None):
        return iter_paths_matching_substr(self.data_dir, substr)

    def iter_experiments(self, substr=None):
        yield from (x.parent for x in self.iter_experiment_entries(substr))

    def iter_experiment_entries(self, substr=None):
        yield from (x for x in iter_paths_matching_substr(
            self.notebook_dir, substr, f'/{8*"[0-9]"}_{{0}}/{{0}}.rst'))

    def iter_notebook_entries(self, substr=None):
        yield from (x for x in iter_paths_matching_substr(
            self.notebook_dir, substr, '/{}.rst'))
        yield from self.iter_experiment_entries(substr)

    def iter_protocols(self, substr=None):
        for dir in self.protocols_dirs:
            yield from self.iter_protocols_from_dir(dir, substr)

    def iter_protocols_from_dir(self, dir, substr=None):
        from . import readers
        extensions = readers.get_known_extensions()
        if substr is not None: extensions.add('')

        include = ['{}' + x for x in extensions]
        exclude = [
                f'**/{8*"[0-9]"}_*',  # Exclude date-stamped protocols.
        ]
        yield from iter_paths_matching_substr(dir, substr, include, exclude)

    def pick_path(self, substr, paths, default=None, no_choices=None):
        paths = list(paths)

        # Complain if there are no paths to pick from.
        if len(paths) == 0:
            raise no_choices or CantMatchSubstr('choices', substr)

        # If there's only one option, return it.
        if len(paths) == 1:
            return paths[0]

        # If the user didn't specify anything, just make our best guess and run 
        # with it.
        if substr is None:
            if default is not None:
                return default

            # Sort the paths alphabetically and return the last one, which will 
            # be the most recent if the paths are prefixed by date.
            resolved_paths = [x.resolve() for x in paths]
            resolved_paths.sort()

            return paths[-1]

        # If the user specified a pattern, but there are multiple matches, ask 
        # for clarification.
        #
        # Once I've written the config-file system, there should be an option 
        # to change how this works (i.e. CLI vs GUI vs automatic choice).
        i = utils.pick_one(x.name for x in paths)
        return paths[i]

    def pick_data(self, substr):
        return self.pick_path(
                substr, self.iter_data(substr),
                no_choices=CantMatchSubstr('data files', substr),
        )

    def pick_experiment(self, substr):
        return self.pick_path(
                substr, self.iter_experiments(substr),
                default=self.current_experiment_dir,
                no_choices=CantMatchSubstr('experiments', substr),
        )

    def pick_notebook_entry(self, substr):
        return self.pick_path(
                substr, self.iter_notebook_entries(substr),
                default=self.current_experiment_entry,
                no_choices=CantMatchSubstr('notebook entries', substr),
        )

    def pick_protocol(self, substr):
        return self.pick_path(
                substr, self.iter_protocols(substr),
                no_choices=CantMatchSubstr('protocols', substr),
        )

    def pick_protocol_reader(self, path, args):
        path = Path(path)
        return readers.pick_reader(path, args)

    def init_project(self, title):
        from cookiecutter.main import cookiecutter
        from .cookiecutter import cookiecutter_path
        from click.exceptions import Abort

        url = self.config.get('cookiecutter_url') or cookiecutter_path

        try:
            cookiecutter(str(url), extra_context={
                    'project_title': title,
                    'year': datetime.today().year,
            })
        except Abort:
            raise KeyboardInterrupt

    def init_experiment(self, title):
        slug = slug_from_title(title)
        expt = self.notebook_dir / f'{utils.ymd()}_{slug}'
        rst = expt / f'{slug}.rst'

        if not os.path.isdir(expt):
            expt.mkdir()
        else:
            sys.exit('Experiment exists. Use exmemo [note] edit [<substr>].')

        with rst.open('w') as file:
            file.write(f"""\
{'*' * len(title)}
{title}
{'*' * len(title)}
""")
        self.launch_editor(rst)

    def launch_editor(self, path):
        if platform.system() == 'Windows':
            system_editor = 'write.exe'
        else:
            system_editor = 'vim'

        editor = self.config.get('editor', os.environ.get('EDITOR', system_editor))
        cmd = *shlex.split(editor), str(path)
        subprocess.Popen(cmd)

    def launch_terminal(self, dir):
        term = self.config.get('terminal', os.environ.get('TERMINAL')) or 'xterm'
        cmd = shlex.split(term)
        subprocess.Popen(cmd, cwd=dir, stdout=DEVNULL, stderr=DEVNULL)

    def launch_pdf(self, path):
        pdf = self.config.get('pdf', os.environ.get('PDF')) or 'evince'
        cmd = *shlex.split(pdf), str(path)
        subprocess.Popen(cmd)

    def launch_browser(self, url, new_window=False):
        browser = self.config.get('browser', os.environ.get('BROWSER')) or 'firefox'
        new_window_flag = self.config.get('browser_new_window_flag', os.environ.get('BROWSER_NEW_WINDOW_FLAG')) or '--new-window'

        if new_window:
            cmd = *shlex.split(browser), *shlex.split(new_window_flag), str(url)
        else:
            cmd = *shlex.split(browser), str(url)

        subprocess.Popen(cmd)

    def sync_data(self, verbose):
        from . import collectors
        collectors.sync_data(self, verbose)

    def build_notebook(self, force=False):
        if force:
            make = 'make', 'clean', 'html'
        else:
            make = 'make', 'html'

        subprocess.run(make, cwd=self.notebook_dir)

    def get_notebook_entry(self, dir):
        dir = Path(dir)
        date, slug = dir.name.split('_', 1)
        return dir / f"{slug}.rst"

    def get_data_collectors():
        pass


class WorkspaceNotFound(IOError):
    show_message_and_die = True

    def __init__(self, dir):
        self.message = f"'{dir}' is not a workspace."


class CantMatchSubstr(Exception):
    show_message_and_die = True

    def __init__(self, type, substr):
        self.message = f"No {type} matching '{substr}'."

def slug_from_title(title):

    class Sanitizer: #
        def __getitem__(self, ord): #
            char = chr(ord)
            if char.isalnum(): return char.lower()
            if char in ' _-': return '_'
            # Any other character gets dropped.

    return title.translate(Sanitizer())

def iter_paths_matching_substr(dir, substr=None, include=None, exclude=None, symlinks=True, include_origin=False):
    substr = '*' if substr is None else f'*{substr}*'
    substr = substr.replace('/', '*/**/*')
    include = _parse_globs(include, substr, ['**/{}', '**/{}/**'])
    exclude = _parse_globs(exclude, substr, '.*')

    matches = sorted(
            Path(x) for x in formic.FileSet(
                directory=dir,
                include=include,
                exclude=exclude,
                symlinks=symlinks,
            )
    )
    if include_origin:
        yield from ((dir, x) for x in matches)
    else:
        yield from matches

def _parse_globs(globs, substr, default):
    if globs is None:
        globs = default
    if isinstance(globs, str):
        globs = [globs]

    return [x.format(substr) for x in globs]







#!/usr/bin/env python3

from . import cli
from .. import Workspace
from pprint import pprint

def config():
    """\
    Dump all the configuration values for the current project.

    Usage:
        exmemo debug config

    Each project reads configuration information from three different files.  
    Typical paths for a Linux system are given below:

        In-project: /path/to/project/.exmemorc
        User-wide: ~/.config/exmemo/conf.toml
        Site-wide: /etc/xdg/exmemo/conf.toml

    Values in the earlier files take precedence over those in later files.  All 
    three files are TOML (https://github.com/toml-lang/toml).  As of yet there 
    is no definitive list of every configuration option understood by exmemo, 
    but typically the help message for each command will mention any relevant 
    configuration options.
    """
    args = cli.parse_args_via_docopt()
    work = Workspace.from_cwd(strict=False)

    pprint(work.config_paths)
    pprint(work.config)

def readers():
    """\
    Dump all of the protocol reader plugins that exmemo is aware of.

    Usage:
        exmemo debug readers
    """
    from ..readers import get_readers
    pprint(get_readers())

def collectors():
    """\
    Dump all of the data collector plugins that exmemo is aware of.

    Usage:
        exmemo debug collector
    """
    from ..collectors import get_collectors
    pprint(get_collectors())








#!/usr/bin/env python3

import shlex
import shutil
import wet_copy
import subprocess
from pathlib import Path
from fossilize import fossilize

def get_readers():
    from .plugins import get_plugins
    return get_plugins('exmemo.protocolreaders')

def get_known_extensions():
    return {
            extension
            for Reader in get_readers()
            for extension in Reader.extensions
    }

def pick_reader(path, args):
    plugins = get_readers()

    for plugin in plugins:
        reader = plugin(path, args)
        if reader.can_handle_path():
            return reader

    raise CantReadProtocol(path, args, plugins)

def ymd():
    from datetime import datetime
    now = datetime.now()
    return now.strftime('%Y%m%d')


class Reader:
    extensions = []

    def __init__(self, path, args):
        self.path = Path(path)
        self.args = args
        self.content = None

    def can_handle_path(self):
        return self.path.suffix in self.extensions

    def archive(self, work, dir=None):
        dest = Path(dir or '.') / f'{ymd()}_{self.path.name}'
        shutil.copy(self.path, dest)


class TxtReader(Reader):
    priority = 0
    extensions = ['.txt', '.md', '.rst']

    def show(self, work):
        with self.path.open() as file:
            print(file.read().strip())

    def edit(self, work):
        work.launch_editor(self.path)

    def print(self, work):
        wet_copy.print_protocol(self.path)

    def archive(self, work, dir=None):
        fossilize([self.path], f'{dir or "."}/$.txt')


class ScriptReader(Reader):
    priority = 0
    extensions = {
            '.py': 'python',
            '.sh': 'bash',
    }

    def show(self, work):
        cmd = self.extensions[self.path.suffix], *self.command
        subprocess.run(cmd)

    def edit(self, work):
        work.launch_editor(self.path)

    def print(self, work):
        wet_copy.print_protocol(self.command_str)

    def archive(self, work, dir=None):
        fossilize(self.command, f'{dir or "."}/$.txt')

    @property
    def command(self):
        return (self.path, *self.args)

    @property
    def command_str(self):
        return ' '.join(str(x) for x in self.command)


class DocReader(Reader):
    priority = 0
    extensions = ['.docx', '.doc', '.xlsx', '.xls', '.odt', '.ods']

    def show(self, work):
        cmd = 'libreoffice', self.path
        subprocess.Popen(cmd)

    def edit(self, work):
        raise CantEditProtocol(self)

    def print(self, work):
        office = 'libreoffice', '--headless', '--invisible', '--convert-to-pdf', self.path
        lpr = 'lpr', '-o', 'sides=one-sided', f'{self.path.stem}.pdf'

        subprocess.run(office)
        subprocess.run(lpr)


class PdfReader(Reader):
    priority = 0
    extensions = ['.pdf']

    def show(self, work):
        work.launch_pdf(self.path)

    def edit(self, work):
        raise CantEditProtocol(self)

    def print(self, work):
        lpr = 'lpr', '-o', 'sides=one-sided', f'{self.path}'
        subprocess.run(lpr)



class CantReadProtocol(Exception):
    show_message_and_die = True

    def __init__(self, path, args, plugins):
        self.message = f"""\
Can't read '{' '.join([str(path), *args])}'.
Tried using the following plugins: {', '.join(x.name for x in plugins)}"""


class CantEditProtocol(Exception):
    show_message_and_die = True

    def __init__(self, reader):
        self.message = f"""\
Can't edit '{reader.path}'.
Editing is not supported by the '{reader.name}' plugin."""



#!/usr/bin/env python3.6
# -*- coding: utf-8 -*-
#
# sgRNA sensor documentation build configuration file, created by
# sphinx-quickstart on Wed Mar 29 13:12:09 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.

## General configuration

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '1.5'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [ #
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinxcontrib.programoutput',
    'exmemo.sphinx.notebook',
    'exmemo.sphinx.biology',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['.templates']

# Ignore warnings about duplicate citations.
suppress_warnings = ['ref.citation']

# The suffix(es) of source filenames.
# You can specify multiple suffixes as a list of strings.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = '{{ cookiecutter.project_title }}'
copyright = '{{ cookiecutter.year }}, {{ cookiecutter.full_name }}'
author = '{{ cookiecutter.full_name }}'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = ''
# The full version, including alpha/beta/rc tags.
release = ''

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = [ #
        'build',
        'Thumbs.db',
        '.DS_Store',
]

rst_epilog = ".. |br| raw:: html\n\n   <br />"

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

## Options for HTML output

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
from sphinx_rtd_theme import get_html_theme_path
html_theme = "sphinx_rtd_theme"
html_theme_path = [get_html_theme_path()]

# Specify the favicon.
from exmemo.sphinx import favicon_path
html_favicon = str(favicon_path)

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['.static']

## Options for HTMLHelp output

# Output file base name for HTML help builder.
htmlhelp_basename = '{{ cookiecutter.project_slug }}'

## Options for LaTeX output

latex_elements = { #
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [ #
    (master_doc, '{{ cookiecutter.project_slug }}.tex', '{{ cookiecutter.project_title }}',
     '{{ cookiecutter.full_name }}', 'manual'),
]

## Options for manual page output

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [ #
    (master_doc, '{{ cookiecutter.project_slug }}', '{{ cookiecutter.project_title }}',
     [author], 1)
]

## Options for Texinfo output

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [ #
    (master_doc, '{{ cookiecutter.project_slug }}', '{{ cookiecutter.project_title }}',
     author, '{{ cookiecutter.project_slug }}', 'One line description of project.',
     'Miscellaneous'),
]






#!/usr/bin/env python3

import subprocess
from . import cli
from .. import Workspace, utils
from pathlib import Path

@cli.priority(10)
def sync():
    """\
    Import data into the project from any available source.

    Usage:
        exmemo [data] sync [-v]

    Options:
        -v --verbose
            Print out each command that is run to sync the data.

    You have lots of control over how exmomo finds and imports data.  The basic 
    idea is to define "data collectors" by adding [[data]] sections any of your  
    configuration files.  Each of these sections should contain a type and any 
    options specific to that type, e.g.:

        [[data]]
        type = ...
        this_option = ...
        that_option = ...

    The type specifies the algorithm that will be used to find and import 
    files.  The following types of data collectors are currently installed:

    {installed_collectors}

    If none of these collectors can handle the data you want to sync, it's easy 
    to write your own.  Each collector is just a class that adheres to the 
    following interface:

       {cls} MyCollector:
           \"\"\"
           Description of options...
           \"\"\"
       
           def __init__(self, this_option, that_option):
              pass

           def sync(self, workspace, verbose):
              pass

    The docstring is used to populate this usage document.  The arguments to 
    the constructor define the options that the collector understands.  The 
    sync() method does the actual work of syncing the data.  The workspace 
    argument is an object that contains information about all the paths in the 
    project, and the verbose argument is just a boolean indicating whether or 
    not you should print out every command that gets run.  Once you've written 
    a class that implements this interface, you can register it with the 
    'exmemo.datacollectors' entry point via the setuptools plugin API to make 
    it available for use.
    """
    args = cli.parse_args_via_docopt(
            cls='class',
            installed_collectors=format_collectors_for_docopt(),
    )
    work = Workspace.from_cwd()

    # Add usage text for each collector to the docstring.
    #from ..collectors import get_collectors

    work.sync_data(args['--verbose'])

@cli.priority(10)
def link():
    """
    Make a symbolic link to the indicated data file.

    Usage:
        exmemo [data] link <substr> [<dir>]

    Arguments
        <substr>
            A string specifying the data file to link.  You can provide any 
            substring from the name of the data file.  If the substring is not 
            unique, you'll be asked which file you meant.

        <dir>
            The directory where the link will be created.  By default, this is 
            the current working directory ('.').
            
    This command is most commonly used to link data files into particular 
    experiments in the notebook directory.  Thus the data directory is a 
    repository for all of your data, and the notebook directory is a curated 
    set of only the most important or relevant data.
    """
    args = cli.parse_args_via_docopt()
    work = Workspace.from_cwd()
    data = work.pick_data(args['<substr>'])
    link = Path(args['<dir>'] or '.') / data.name

    link.symlink_to(data)

def gel():
    """
    Format a gel for inclusion in the lab notebook.  In particular, this means 
    converting it from a *.tif to a *.png, inverting it's colors, and giving 
    you a chance to crop it.

    Usage:
        exmemo data gel <substr> [<dir>] [-f]

    Arguments:
        <substr>
            A string specifying the data file to copy and crop.  You can 
            provide any substring from the name of the data file.  If the 
            substring is not unique, you'll be asked which file you meant.

        <dir>
            The directory where the image will be copied.  By default, this is 
            the current working directory ('.').

    Options:
        -f, --force
            Overwrite any existing image.

    This command is commonly used to copy gels into the notebook directory.  
    The full resolution images stay in the data directory, but you can make a 
    nicely formatted copy to show in the notebook.
    """
    args = cli.parse_args_via_docopt()
    work = Workspace.from_cwd()
    data = work.pick_data(args['<substr>'])
    copy = Path(args['<dir>'] or '.') / f'{data.stem}.png'

    if copy.exists() and not args['--force']:
        raise FileAlreadyExists(copy)

    convert = 'convert', str(data), '-fill', 'white', '-opaque', 'red', '-negate', str(copy)
    gimp = 'gimp', str(copy)

    subprocess.run(convert)
    subprocess.Popen(gimp)

def ls():
    """\
    List data files.

    Usage:
        exmemo data ls [<substr>]

    Arguments:
        <substr>
            Only list files that contain the given substring.
    """
    args = cli.parse_args_via_docopt()
    work = Workspace.from_cwd()

    for path in work.iter_data(args['<substr>']):
        print(path.relative_to(work.data_dir))


def format_collectors_for_docopt():
    from textwrap import indent
    from ..collectors import get_collectors

    collectors = get_collectors()
    collectors = sorted(collectors.values(), key=lambda x: x.name)

    usage = ""

    for last, collector in utils.last(collectors):
        docstring = cli.get_docstring(collector)
        usage += f"'{collector.name}'" + '\n'
        usage += indent(docstring, '    ')
        if not last: usage += '\n\n'

    return usage

class FileAlreadyExists(Exception):
    show_message_and_die = True

    def __init__(self, path):
        self.message = f"Refusing to overwrite {path}"






#!/usr/bin/env python3
# encoding: utf-8

from setuptools import setup

with open('README.rst') as file:
    readme = file.read()

setup(
    name='exmemo',
    version='0.2.0',
    author='Kale Kundert',
    author_email='kale@thekunderts.net',
    url='https://github.com/kalekundert/exmemo',
    long_description=readme,
    include_package_data=True,
    packages=[
        'exmemo',
        'exmemo.commands',
        'exmemo.cookiecutter',
        'exmemo.sphinx',
    ],
    install_requires=[
        'appdirs',
        'cookiecutter',
        'docopt',
        'formic-py3',
        'fossilize',
        #'metapub',
        'habanero',
        'nonstdlib',
        'PyYAML',
        'sphinx',
        'sphinxcontrib-programoutput',
        'sphinx_rtd_theme',
        'toml',
        'wet_copy',
    ],
    entry_points={
        'console_scripts': [
            'exmemo=exmemo.commands.main:main',
        ],
        'exmemo.commands': [
            'init=exmemo.commands.project:init',
            'new=exmemo.commands.note:new',
            'edit=exmemo.commands.note:edit',
            'open=exmemo.commands.note:open',
            'build=exmemo.commands.note:build',
            'browse=exmemo.commands.note:browse',
            'show=exmemo.commands.protocol:show',
            'print=exmemo.commands.protocol:print',
            'archive=exmemo.commands.protocol:archive',
            'sync=exmemo.commands.data:sync',
            'link=exmemo.commands.data:link',
            'project=exmemo.commands.main:project',
            'note=exmemo.commands.main:note',
            'protocol=exmemo.commands.main:protocol',
            'data=exmemo.commands.main:data',
            'config=exmemo.commands.main:config',
            'debug=exmemo.commands.main:debug',
        ],
        'exmemo.commands.project': [
            'init=exmemo.commands.project:init',
            'root=exmemo.commands.project:root',
        ],
        'exmemo.commands.note': [
            'new=exmemo.commands.note:new',
            'edit=exmemo.commands.note:edit',
            'open=exmemo.commands.note:open',
            'directory=exmemo.commands.note:directory',
            'build=exmemo.commands.note:build',
            'browse=exmemo.commands.note:browse',
            'ls=exmemo.commands.note:ls',
        ],
        'exmemo.commands.data': [
            'sync=exmemo.commands.data:sync',
            'link=exmemo.commands.data:link',
            'gel=exmemo.commands.data:gel',
            'ls=exmemo.commands.data:ls',
        ],
        'exmemo.commands.protocol': [
            'show=exmemo.commands.protocol:show',
            'edit=exmemo.commands.protocol:edit',
            'print=exmemo.commands.protocol:print',
            'archive=exmemo.commands.protocol:archive',
            'plugins=exmemo.commands.protocol:plugins',
            'ls=exmemo.commands.protocol:ls',
        ],
        'exmemo.commands.config': [
            'get=exmemo.commands.config:get',
            'set=exmemo.commands.config:set',
            'edit=exmemo.commands.config:edit',
        ],
        'exmemo.commands.debug': [
            'config=exmemo.commands.debug:config',
            'readers=exmemo.commands.debug:readers',
            'collectors=exmemo.commands.debug:collectors',
        ],
        'exmemo.protocolreaders': [
            'txt=exmemo.readers:TxtReader',
            'exe=exmemo.readers:ScriptReader',
            'doc=exmemo.readers:DocReader',
            'pdf=exmemo.readers:PdfReader',
        ],
        'exmemo.datacollectors': [
            'rsync=exmemo.collectors:RsyncCollector',
            'usb=exmemo.collectors:UsbCollector',
        ],
    },
    classifiers=[
            'Topic :: Scientific/Engineering',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 3.6',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Development Status :: 2 - Pre-Alpha',
    ],
)
